# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.
#
# Copyright (c) 2016-2022 Intra2net AG <info@intra2net.com>

"""
test_binary.py: unit tests for cnfvar/binary.py.

Tests classes and functions in cnfvar/binary.py

For help see :py:mod:`unittest`
"""

import unittest
from textwrap import dedent
from types import SimpleNamespace
from unittest.mock import patch, ANY

from src.cnfvar import binary
from src.cnfvar.binary import CnfBinary


class CnfBinaryTest(unittest.TestCase):
    """Test class CnfBinary."""

    #: return value for the run method
    CMD_STDOUT = ""
    CMD_STDERR = ""

    def setUp(self):
        """Set up mocks."""
        self.CMD_STDOUT = ""
        self.CMD_STDERR = ""
        self._proc_patch = patch.object(binary.subprocess, "run", side_effect=self._fake_subprocess_run)
        self._proc_mock = self._proc_patch.start()

    def tearDown(self):
        """Clean up mocks."""
        self._proc_patch.stop()
        self._proc_mock = None
        self.CMD_STDOUT = ""
        self.CMD_STDERR = ""

    def test_basic_get_cnf(self):
        """Tests basic functionality of the get_cnf wrapper."""
        self.CMD_STDOUT = dedent("""\
            1 USER,1: "jake"
            2    (1) USER_DISABLED,0: "0"
            3    (1) USER_FULLNAME,0: "Jake"
            4    (1) USER_GROUPWARE_FOLDER_DRAFTS,0: "INBOX/Entwürfe"
            5    (1) USER_GROUPWARE_FOLDER_OUTBOX,0: "INBOX/Gesendete Objekte"
            6    (1) USER_GROUPWARE_FOLDER_TRASH,0: "INBOX/Gelöschte Elemente"
            7    (1) USER_GROUP_MEMBER_REF,0: "100"
            8    (1) USER_GROUP_MEMBER_REF,1: "2"
            9    (1) USER_PASSWORD,0: "test1234"
            11 USER,2: "jill"
            12    (11) USER_DISABLED,0: "0"
            13    (11) USER_FULLNAME,0: "Jill"
            14    (11) USER_GROUPWARE_FOLDER_DRAFTS,0: "INBOX/Entwürfe"
            15    (11) USER_GROUPWARE_FOLDER_OUTBOX,0: "INBOX/Gesendete Objekte"
            16    (11) USER_GROUPWARE_FOLDER_TRASH,0: "INBOX/Gelöschte Elemente"
            17    (11) USER_GROUP_MEMBER_REF,0: "100"
            18    (11) USER_GROUP_MEMBER_REF,1: "2"
            19    (11) USER_PASSWORD,0: "test1234"
            """)

        output = CnfBinary.get_cnf("USER", 10)
        self.assertEqual(self.CMD_STDOUT.splitlines(), output.splitlines())
        self._proc_mock.assert_called_once_with(["/usr/intranator/bin/get_cnf", "USER", "10"],
                                                input=ANY, check=ANY, capture_output=ANY,
                                                encoding='latin1', timeout=ANY)

    def test_get_cnf_error_handling(self):
        """Test that the get cnf method does some sanity checks."""
        # cannot pass an instance without a name
        with self.assertRaises(ValueError):
            CnfBinary.get_cnf(None, 10)

        # cannot pass an instance of an invalid type
        with self.assertRaises(TypeError):
            CnfBinary.get_cnf("USER", 18.45)

    def test_basic_set_cnf(self):
        """Test basic set cnf operation."""
        input_str = """\
            1  FIREWALL_SERVICEGROUP,1: "http"
            2     (1) FIREWALL_SERVICEGROUP_PREDEFINED_ID,0: "1"
            3     (1) FIREWALL_SERVICEGROUP_TYPE,0: "TCP"
            4        (3) FIREWALL_SERVICEGROUP_TYPE_COMMENT,0: ""
            5        (3) FIREWALL_SERVICEGROUP_TYPE_TCPUDP_DST_PORT,0: "80"
        """

        CnfBinary.set_cnf(input_str)
        self._proc_mock.assert_called_once_with(
            ["/usr/intranator/bin/set_cnf"],
            input=input_str, check=ANY, capture_output=ANY,
            encoding="latin1", timeout=ANY)

    def test_set_cnf_extra_args(self):
        """Test that we can pass extra arguments to the set_cnf binary."""
        input_str = """\
        {
            "cnf" :
            [
                {
                    "data" : "0",
                    "instance" : 0,
                    "number" : 1,
                    "varname" : "ACME_DEBUG_ENABLE"
                }
            ]
        }
        """

        CnfBinary.set_cnf(input_str, as_json=True, fix_problems=True, discard_queue=True)
        self._proc_mock.assert_called_once_with(
            ["/usr/intranator/bin/set_cnf", "--json", "--fix-problems", "--discard-queue"],
            input=input_str, check=ANY, capture_output=ANY,
            encoding="utf8", timeout=ANY)

    def _fake_subprocess_run(self, *args, **kwargs):
        return SimpleNamespace(args=args[0], stdout=self.CMD_STDOUT,
                               stderr=self.CMD_STDERR, returncode=0)


if __name__ == '__main__':
    unittest.main()
